@@ -14,10 +14,13 @@ from Bio.Alphabet import IUPAC
 
 import ProteinObject
 import Networking
+import httplib
+import Parser as GRP
+import UniprotAPI
 
 ######################################################### 
 #########################################################
-class ProteinRequestParser:
+class ProteinRequestParser(GRP.GeneralRequestParser):
 
 # Initialization function, set Entrez.email for calls, an ensure key -1 is set
 # to a non-existant object
@@ -25,19 +28,31 @@ class ProteinRequestParser:
     def __init__(self, email, cache, retry=0, loud=True):
         """"Initializes an empty requestParser object, setting the Entrez.email field and defining how many times network errors should be retried"""
         try:
-            Entrez.email = email
-            self.Networking = Networking.Networking(30)
-            self.loud = loud
-            self.retry = retry
-            self.cache = cache
-            self.protein_datastore = {-1 : ProteinObject.ProteinObject([])}
+            GRP.GeneralRequestParser.__init__(self, email, cache, retry, loud)
+        
+            self.protein_datastore = {-1 : ProteinObject.ProteinObject(-1, [])}
             self.protein_translationMap = {-1: -1}
-            self.error_status = False
             self.batchableFunctions = [self.get_sequence, self.get_protein_name, self.get_variants, self.get_geneID, self.get_protein_sequence_length]
-           
-        except: 
+            self.UniprotAPI = UniprotAPI.UniprotAPI()
+            
+            self.error_status = False
+             
+        except Exception, e: 
             print "Fatal error when creating ProteinRequestParserObject"
             self.error_status = True
+            raise e
+
+
+    def has_key(self, ID):
+        return self.protein_datastore.has_key(ID)
+
+    def keys(self):
+        kl = []
+        for i in self.protein_datastore:
+            kl.append(i)
+
+        kl.remove(-1)
+        return kl
             
 #--------------------------------------------------------
 # PUBLIC FUNCTION
@@ -98,17 +113,61 @@ class ProteinRequestParser:
         return ProtObj.get_protein_sequence_length()
 
 
+#--------------------------------------------------------
+# PUBLIC FUNCTION
+#--------------------------------------------------------
+#   
+# Get the sequence length of the protein's AA sequence
+    def get_other_accessions(self, ID):
+        ProtObj = self._get_protein_object(ID)
+        return ProtObj.get_other_accessions()
+
+
+#--------------------------------------------------------
+# PUBLIC FUNCTION
+#--------------------------------------------------------
+#   
+# Get the sequence length of the protein's AA sequence
+    def get_species(self, ID):
+        ProtObj = self._get_protein_object(ID)
+        return ProtObj.get_species()
+
+#--------------------------------------------------------
+# PUBLIC FUNCTION
+#--------------------------------------------------------
+#   
+# Get the sequence length of the protein's AA sequence
+    def get_taxonomy(self, ID):
+        ProtObj = self._get_protein_object(ID)
+        return ProtObj.get_taxonomy()
+
+#--------------------------------------------------------
+# PUBLIC FUNCTION
+#--------------------------------------------------------
+#   
+# Get the sequence length of the protein's AA sequence
+    def get_domains(self, ID):
+        ProtObj = self._get_protein_object(ID)
+        return ProtObj.get_domains()
+
+#--------------------------------------------------------
+# PUBLIC FUNCTION
+#--------------------------------------------------------
+#   
+# Get the sequence length of the protein's AA sequence
+    def get_gene_name(self, ID):
+        ProtObj = self._get_protein_object(ID)
+        return ProtObj.get_gene_name()
+
 #--------------------------------------------------------
 # PUBLIC FUNCTION
 #--------------------------------------------------------
 #
-# TO DO - get linear AA sequence distance
-
-    def get_distance_between_residues(self, ID, R1, R2):
-        print "!!!! NOT YET IMPLEMENTED !!!!"
-        return 0;
-
+# Get any isoforms associated with this sequence ID
 
+    def get_isoforms(self, ID):
+        ProtObj = self._get_protein_object(ID)
+        return ProtObj.get_isoforms()
 #--------------------------------------------------------
 # PUBLIC FUNCTION
 #--------------------------------------------------------
@@ -170,7 +229,6 @@ class ProteinRequestParser:
         return ID_type(ProteinID)
 
 
-
 #--------------------------------------------------------
 # PUBLIC FUNCTION
 #--------------------------------------------------------
@@ -178,7 +236,7 @@ class ProteinRequestParser:
 #
     def purge_data_store(self):
         del self.protein_datastore
-        self.protein_datastore = {-1 : ProteinObject.ProteinObject([])}
+        self.protein_datastore = {-1 : ProteinObject.ProteinObject(-1, [])}
 
 #--------------------------------------------------------
 # PUBLIC FUNCTION
@@ -188,127 +246,9 @@ class ProteinRequestParser:
     def get_size_of_datastore(self):
         return len(self.protein_datastore)-1
 
-
-#--------------------------------------------------------
-# PRIVATE FUNCTION
-#--------------------------------------------------------
-#
-# internal function which returns a protein object either from
-# the cache (if caching is on and the protein data has already 
-# been downloaded) or directly from the NCBI database if caching
-# is off or the protein isn't yet in the database.
-#
-    def _get_protein_object(self, proteinID):
-       
-        if proteinID not in self.protein_datastore or not self.cache:  
-
-            protein_xml = -1
-            retry = self._build_retry_function(self.Networking.efetchProtein);
-            
-            # run conversion if necessary
-            proteinID = self._convertIfNecessary(proteinID)
-
-            # if we can be sure this type of ID will not return a protein
-            # because its an invald accession number
-            if ID_type(proteinID)[0] == -1:
-
-                self.printWarning("\nWarning - The ID {ID} is an invalid accession number, and the database will not be queried".format(ID=proteinID))
-                # by returning the object associated with [] we don't pollute the datastore with invalid and pointless
-                # searches, we avoid queriying NCBI without a hope in hell of a hit, and we take advantage of the built in
-                # bad XML behaviour without raising an error, because, technically, no error has happened, we just know
-                # that the ID in question won't return protein data. It's not an error - it's just stupid.
-            
-                return ProteinObject.ProteinObject([])
-            
-            # ---------------------------------------------------------------------------------
-            # check if handle represents an error
-            # PLEASE NOTE that if you use a protein ID which fails to get a protein
-            # this is *not* an error (and Entrez.efetch will behave accordingly), so
-            # instead in such a situtation we simply return an empty ProteinObject with
-            # the exists attribute set to False, but Error is also false.
-            # IF, however, protein_handle returns -1 (indicating some error) then we return
-            # an empty ProteinObject with error set to True
-            # ---------------------------------------------------------------------------------            
-          
-            while (protein_xml == -1):
-                protein_xml = retry(proteinID);
-                        
-            # if we still can't get through after retrying a number of times
-            if (protein_xml == -2):
-                print "Unable to find accessionValue at NCBI end"
-                self.protein_datastore[proteinID] = ProteinObject.ProteinObject(-1)
-               
-            else:
-                self.protein_datastore[proteinID] = ProteinObject.ProteinObject(protein_xml)
-                
-        return self.protein_datastore[proteinID]
-
-#--------------------------------------------------------
-# PRIVATE FUNCTION
-#--------------------------------------------------------          
-# Function which we can manually add cases to where converting
-# from some kind of accession number to a GI would be better
-# for whatever reason. Currently only true for PDB IDs
-#
-
-    def _convertIfNecessary(self, ProteinID):
-        
-        # if PDB
-        if ID_type(ProteinID)[0] == 6:
-            return self.translate_Asc2GI(ProteinID)
-        
-        return ProteinID
-
-#--------------------------------------------------------
-# PRIVATE FUNCTION
-#--------------------------------------------------------          
-# Builds a closure based retry function, which when called
-# the first self.retry times will atempt to get the parsed
-# xml for the protein ID in question. However, on the
-# self.retry+1 time it will simply return -2 
-#
-    def _build_retry_function(self, function_to_apply):
-
-        retryCounter = [0]
-        numberOfRetries = self.retry
-
-        def retry(ProteinID):
-
-            if retryCounter[0] < numberOfRetries+1:
-                
-                retryCounter[0] = retryCounter[0]+1
-                
-                ## if we're not on our first try
-                if not retryCounter[0]-1 == 0:
-                    print("Retry number " + str(retryCounter[0]) + " of " + str(numberOfRetries+1))
-                
-                ## return value may be a real handle or -1
-                handle = function_to_apply(ProteinID)
-                
-                ## if we failed return -1
-                if handle == -1:
-                    return -1
-                
-                try:
-                    proteinXML = Entrez.read(handle)
-                except (Bio.Entrez.Parser.CorruptedXMLError, Bio.Entrez.Parser.NotXMLError, Bio.Entrez.Parser.ValidationError), err:      
-                    return -1
-
-                return proteinXML
-
-            else:
-                return -2;
-
-        return retry
-
-
 #--------------------------------------------------------
 # PUBLIC FUNCTION
 #--------------------------------------------------------
-# Function which takes one of the get_? functions defined
-# below and an array of IDS of interest, and returns 
-# tuple of ID-results while remaining INSIDE NCBI's
-# fetch parameters.
 #
 
     def batchFetch(self, function, listOfIDs):
@@ -327,7 +267,7 @@ class ProteinRequestParser:
         for proteinID in listOfIDs:
             if proteinID not in self.protein_datastore or not self.cache:
                 proteinID = self._convertIfNecessary(proteinID)
-                if not ID_type(proteinID)[0] == -1:
+                if not ID_type(proteinID)[0] < 0:
                     toFetch.append(proteinID)                    
                            
         # next we take those which are NOT in the datastore and take advantage
@@ -335,68 +275,67 @@ class ProteinRequestParser:
         # call to the server, so is a lot faster (reduces setup and teardown)
         # generates a list, each element of which is the 1:1 xml for the 
         # listOfIDsF
-        listOfXML = self._get_batch_XML(toFetch)
-            
+        listOfXML = self._get_batch_XML(toFetch, self.Networking.efetchProtein, self.UniprotDatabaseLookup)
+        
         # note we don't have to try/catch here because _get_batch_XML() guarentees
         # that each XML field is valid
         fetchCounter = 0
         
         for protein_xml in listOfXML:
-            self.protein_datastore[toFetch[fetchCounter]] = ProteinObject.ProteinObject([protein_xml])
+            if not self.protein_datastore.has_key(toFetch[fetchCounter]) and not protein_xml == -1 :
+                self.protein_datastore[toFetch[fetchCounter]] = ProteinObject.ProteinObject(toFetch[fetchCounter], [protein_xml])
+            
             fetchCounter = fetchCounter+1
+
                         
         # finally, we build a tuple with results from everything in the input, and
         # return. Note we're ONLY doing this if that ID has already been loaded into
         # the data store. This stops the retry mechanism going through for batch, failing
         # and then just retrying again via the _get_protein_object() function which the
         # $function inevitably will call
+        #
+        # By checking against the datastore we also ensure we can pick up additional records
+        # built by the "alternative" function, which adds to the datastore in an independent manner
+        # to the for loop above 
+            
         for ID in listOfIDs:
             if ID in self.protein_datastore:
                 outputList[ID] = function(ID)
+                
             else:
                 outputList[ID] = function(-1)
 
         return outputList
 
+    def _get_protein_object(self, proteinID):
+
+        proteinID = self._convertIfNecessary(proteinID)
+
+        # pre-check to see if the accession matches the predefined accession format
+        # rules. If it doesn't then 
+        if ID_type(proteinID)[0] < 0:
+            self.printWarning("\nWarning - The ID {ID} is an invalid accession number, and the database will not be queried".format(ID=proteinID))
+            proteinID = -1
+        else:
+            self._get_object(proteinID, self.protein_datastore, self.Networking.efetchProtein, ProteinObject.ProteinObject, self.UniprotDatabaseLookup)
+        
+        return self.protein_datastore[proteinID]
+
 
 #--------------------------------------------------------
 # PRIVATE FUNCTION
-#-------------------------------------------------------
-# This function provides an interface with the Networking
-# functionality to return a list of XML elements, each of
-# which corresponds to the ID in the listofIDs
+#--------------------------------------------------------          
+# Function which we can manually add cases to where converting
+# from some kind of accession number to a GI would be better
+# for whatever reason. Currently only true for PDB IDs
 #
-#
-    def _get_batch_XML(self, listOfIDs):
+
+    def _convertIfNecessary(self, ProteinID):
+        # if PDB
+        if ID_type(ProteinID)[0] == 6:
+            return self.translate_Asc2GI(ProteinID)
         
-        # base case 1 (base case 2 if len == 1)
-        if len(listOfIDs) == 0:
-            return []
-
-        protein_xml = -1
-        retry = self._build_retry_function(self.Networking.efetchProtein);
-
-        while (protein_xml == -1):
-            protein_xml = retry(listOfIDs)
-
-        # If we fail implement recursive batch fetch, such that we split the list in half
-        # and batch fetch each half. Do this after troubleshooting!
-        if protein_xml == -2 or not len(protein_xml) == len(listOfIDs):
-            if len(listOfIDs) == 1:
-                return [-1]
-            
-            # split list in half and recursivly batch both halves
-            # this serves two purposes - it lets us retry a number of times proportional
-            # to the length of the list, and it provides a binary search mechanism to 
-            # root out a potential rotten apple which may be causing the list to error
-
-            protein_xml = self._get_batch_XML(listOfIDs[:int(len(listOfIDs)/2)])
-            protein_xml_2 = self._get_batch_XML(listOfIDs[int(len(listOfIDs)/2):])
-
-            protein_xml.extend(protein_xml_2)
-            
-        return protein_xml
-
+        return ProteinID
 
 #--------------------------------------------------------
 # PRIVATE FUNCTION
@@ -409,8 +348,36 @@ class ProteinRequestParser:
             print message
 
 
+#--------------------------------------------------------
+# PRIVATE FUNCTION
+#--------------------------------------------------------
+# Function which gives the GeneralRequestParser access to
+# non NCBI based means to get a protein record (Uniprot)
+#
+
+    def UniprotDatabaseLookup(self, accessionID):
         
+        IDtype = ID_type(accessionID)[0]
+              
+        # we only query UniProt if the accession registers as a SWISSPROT or UniProt
+        # value
+        if IDtype == 2 or IDtype == 7:
+            print "[UniProt]: Falling back and querying UniProt servers...  "
         
+            # query the server through the UniprotAPI class
+            self.UniprotAPI.getProteinObjectFromUniProt(self.protein_datastore, accessionID)
+                    
+            # NOTE this is crucial - we only return true if we can get an object 
+            # from the dictionary
+            try:
+                self.protein_datastore[accessionID]
+                print "[UniProt]: Sucess!"
+                return True
+            except KeyError:
+                return False
+        else:
+            return False
+
 # +-------------------------------------------------------+
 # |                    END OF CLASS                       |
 # +-------------------------------------------------------+
@@ -422,15 +389,39 @@ class ProteinRequestParser:
 ## ProteinParser function
 ## General function true to the module, as does not require
 ## a class instance to work
+#
+# List of for refrence/addition
+# -2 : IPI - IPI is -2 because we can't look it up, but it 
+#            is a valid type of accession. Neither NCBI nor
+#            Uniprot support IPI lookups, however!
+# -1 : unknown
+#  0 : GI
+#  1 : refseq
+#  2 : Swissprot
+#  3 : DDBJ
+#  4 : genbank
+#  5 : EMBL
+#  6 : PDB
+#  7 : Uniprot
+
 def ID_type(ProteinID):
     
-    # convert ID to all uppercase characters...
+    # convert ID to all uppercase characters, and remove any periods 
     try:
         ProteinID = ProteinID.upper()
     except AttributeError:
         return ([-1, "Unknown protein accession type"])
-        
 
+
+    # cut off . and anything after
+    # "." represent versions of accession numbers, but in this context 
+    # the versions a) don't change the type and b) just make all the
+    # regexes way harder!
+
+    if ProteinID.find(".") > 0:
+        ProteinID = ProteinID[:ProteinID.find(".")]
+
+        
     # if it begin [A|N|X|Y|Z]P_ then it's a refseq 
     if re.match("[ANXYZ][P]_", ProteinID):
         return [1, "RefSeq"]
@@ -439,6 +430,7 @@ def ID_type(ProteinID):
     if ProteinID.isdigit() or re.match("GI",ProteinID):
         return [0, "GI"]
 
+    # is it a PDB?
     if re.match("[0-9][A-Z0-9][A-Z0-9][A-Z0-9]", ProteinID) or re.match("[0-9][A-Z0-9][A-Z0-9][A-Z0-9]_[A-Z0-9]", ProteinID):
         return[6, "PDB"]
     
@@ -446,6 +438,13 @@ def ID_type(ProteinID):
     if re.match("[OPQ]", ProteinID) and re.match("^[A-Z0-9]+$", ProteinID):
         return [2, "Swissprot"]
 
+    # uniprot format - NB: NOT FOUND IN ENTREZ/NCBI DATABASE - will 
+    # have to fall back to UniProt servers to get information
+    if re.match("[A-N|R-Z][0-9][A-Z0-9][A-Z0-9][A-Z0-9][0-9]", ProteinID):
+        return [7, "UniProt"]
+
+    if re.match("IPI[0-9]*", ProteinID):
+        return [-2, "International Protein Index"]
     
     # Now we've removed refseq, gi and swissprot we can be a bit more discerening about 
     # what the accession should be (as now it must conform to the NCBI accesison number
@@ -463,9 +462,7 @@ def ID_type(ProteinID):
     # DDBJ
     if re.match("[B][A-Z][A-Z]", ProteinID) or re.match("F[A-Z][A-Z]", ProteinID) or \
             re.match("G[A-Z][A-Z]", ProteinID) or re.match("I[A-Z][A-Z]", ProteinID):
-        return [3, "DDBJ"]
-
-    
+        return [3, "DDBJ"]    
 
     # GenBank
     if re.match("A[A-Z][A-Z]", ProteinID) or re.match("AAE", ProteinID) or \
